/* CS 472/672 Assignment #2 - dirsolver.h file */

#ifndef DIRSOLVER_H    /* Include guards */
#define DIRSOLVER_H

/*                                                    */
/* Sparse row format data structure for the LU matrix */
/*                                                    */
/*   
   Each row of the LU matrix is held as a sparse_row structure
   for k=0, ..., nz-1
      jaf[k] = col index
      af[k]  = value
      k = diag is the location of the diagonal of the row  
          i.e. if k = row.diag => af[k] = diag entry 
*/

struct sparse_row {
   double *af;        /* Coefficient, as with a */
   int    *jaf;       /* Column index, as with ja */
   int    nz;         /* Number of nonzeros in the row */
   int    diag;       /* Where the diagonal coefficient is stored */
};


/*                       */
/* Numeric factorization */
/*                       */

void factor(int N, int *ia, int *ja, double *a, int *order, int *invord, struct sparse_row *LU);
/*
   input:

      N       number of unknowns
      ia, ja  stucture of a
      a       real values of a
      order   order[new_order] = old_order
      invord  invord[old_order] = new_order
      LU      contains info about symbolic factors of the LU

   output:

      LU      now contains real LU as well as symbolic
*/


/*             */
/* Solve phase */
/*             */

void solve(int N, double *x, double *b, int *order, struct sparse_row *LU);
/*
   input:

      N      number of unknowns
      b      rhs
      order  order[new_order] = old_order
      LU     sparse_row structure containing LU

   output:

      x      solution in the original order
*/


/*                                       */
/* Symbolic factorization                */
/* (merge_smart is called by sfac_smart) */
/*                                       */

void  merge_smart(int i, int *list, int N, int *link, struct sparse_row *LU);
/*
   input:

      i     current row being factored (Dolittle form)
      list  implied linked list of nonzeros in current
            row.  On entry, only original nonzero columsn
            in this row.  1st nonzero column = first,
            list[first], list[ list[first]], ...., until
            list[  [....] ] = N+1 -> signals end of list
      N     number of unknowns
      LU    ptr to sparse row structure containing info about
            previous rows

      link  implied linked list of rows whose first non-zero is in col i, i.e.
            list of prev rows whose first nonzero col is in col i given by
            link[i], link[link[i]], ..., until link[  [....] ] = N+1
            -> signals end of list

   output:

     list   now contains all nonzeros in this factored row i
            (implied linked list)
*/

void sfac_smart(int *ia, int *ja, int N, int *order, int *invord, int *nnz, struct sparse_row *LU);
/*
   input:

     ia, ja  structure of a
     N       number of unknowns
     order   order[new_order] = old_order
     invord  invord[old_order] = new_order

  output:

     nnz     number of nonzeros in LU
     LU      sparse_row stucture contains symbolic ILU info
*/

/* Auxilliary functions */
void sortq(int *v, int left, int right);
void swap(int *v, int i, int j);
void clean(int N, struct sparse_row *LU);

#endif

